#!/usr/bin/env python3
"""Converts mel spectrograms to WAV audio using griffin-lim"""
import argparse
import io
import logging
import os
import sys
import time
from pathlib import Path

import numpy as np
import scipy.io.wavfile

from .audio import AudioSettings
from .utils import add_audio_settings

_LOGGER = logging.getLogger("wav2mel.griffin_lim")


def main():
    """Main entry point"""
    parser = argparse.ArgumentParser(prog="wav2mel.griffin_lim")
    parser.add_argument("--output-dir", help="Directory to write WAV files")
    parser.add_argument("--iterations", type=int, default=60)
    parser.add_argument("mel", nargs="*", help="Mel spectrogram files (.npy)")
    parser.add_argument("--dtype", default="float32", help="numpy data type for mel")
    parser.add_argument(
        "--batch-dim", action="store_true", help="Spectrograms include batch dimension"
    )

    add_audio_settings(parser)

    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to the console"
    )
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    _LOGGER.debug(args)

    # Convert to paths
    if args.output_dir:
        args.output_dir = Path(args.output_dir)
        args.output_dir.mkdir(parents=True, exist_ok=True)

    # -------------------------------------------------------------------------

    audio_settings = AudioSettings(
        # STFT
        filter_length=args.filter_length,
        hop_length=args.hop_length,
        win_length=args.win_length,
        mel_channels=args.mel_channels,
        sample_rate=args.sample_rate,
        mel_fmin=args.mel_fmin,
        mel_fmax=args.mel_fmax,
        ref_level_db=args.ref_level_db,
        spec_gain=args.spec_gain,
        #
        # Normalization
        signal_norm=not args.no_normalize,
        min_level_db=args.min_level_db,
        max_norm=args.max_norm,
        clip_norm=not args.no_clip_norm,
        symmetric_norm=not args.asymmetric_norm,
    )

    # -------------------------------------------------------------------------

    def process_mel(mel_db: np.ndarray, utt_id: str = ""):
        if args.batch_dim:
            mel_db = mel_db.squeeze(0)

        # Run griffin-lim
        _LOGGER.debug("Mel shape: %s", mel_db.shape)

        wav = audio_settings.mel2wav(mel_db, num_iters=args.iterations)
        duration_sec = len(wav) / audio_settings.sample_rate

        # Save WAV data
        if not utt_id:
            # Use timestamp
            utt_id = str(time.time())

        if args.output_dir:
            # Write to file
            wav_path = args.output_dir / (utt_id + ".wav")
            with open(wav_path, "wb") as wav_file:
                scipy.io.wavfile.write(wav_file, audio_settings.sample_rate, wav)

            _LOGGER.debug(
                "Wrote %s (%s sample(s), %s second(s))",
                wav_path,
                len(wav),
                duration_sec,
            )
        else:
            # Write to stdout
            with io.BytesIO() as wav_file:
                scipy.io.wavfile.write(wav_file, audio_settings.sample_rate, wav)
                sys.stdout.buffer.write(wav_file.getvalue())

            _LOGGER.debug("Wrote (%s sample(s), %s second(s))", len(wav), duration_sec)

    # -------------------------------------------------------------------------

    if args.mel:
        for mel_path in args.mel:
            # Load from numpy file
            _LOGGER.debug("Processing %s", mel_path)
            mel_db = np.load(mel_path, allow_pickle=True).astype(args.dtype)
            process_mel(mel_db, utt_id=Path(mel_path).stem)
    else:
        if os.isatty(sys.stdin.fileno()):
            print("Reading mel from stdin...", file=sys.stderr)

        # stdin is single numpy array
        with io.BytesIO(sys.stdin.buffer.read()) as mel_file:
            mel_db = np.load(mel_file, allow_pickle=True).astype(args.dtype)

        process_mel(mel_db)


# -----------------------------------------------------------------------------


if __name__ == "__main__":
    main()



"""Audio settings for wav2mel"""
import typing
from dataclasses import dataclass

import librosa
import numpy as np


@dataclass
class AudioSettings:
    """Settings for wav <-> mel"""

    # STFT settings
    filter_length: int = 1024
    hop_length: int = 256
    win_length: int = 256
    mel_channels: int = 80
    sample_rate: int = 22050
    mel_fmin: float = 0.0
    mel_fmax: typing.Optional[float] = 8000.0
    ref_level_db: float = 20.0
    spec_gain: float = 1.0

    # Normalization
    signal_norm: bool = True
    min_level_db: float = -100.0
    max_norm: float = 4.0
    clip_norm: bool = True
    symmetric_norm: bool = True

    def __post_init__(self):
        if self.mel_fmax is not None:
            assert self.mel_fmax <= self.sample_rate // 2

        # Compute mel bases
        self._mel_basis = librosa.filters.mel(
            self.sample_rate,
            self.filter_length,
            n_mels=self.mel_channels,
            fmin=self.mel_fmin,
            fmax=self.mel_fmax,
        )

        self._inv_mel_basis = np.linalg.pinv(self._mel_basis)

    # -------------------------------------------------------------------------
    # Mel Spectrogram
    # -------------------------------------------------------------------------

    def wav2mel(
        self, wav: np.ndarray, trim_silence: bool = False, trim_db: float = 60.0
    ) -> np.ndarray:
        if trim_silence:
            wav = self.trim_silence(wav, trim_db=trim_db)

        linear = self.stft(wav)
        mel_amp = self.linear_to_mel(np.abs(linear))
        mel_db = self.amp_to_db(mel_amp)

        if self.signal_norm:
            mel_db = self.normalize(mel_db)

        return mel_db

    def mel2wav(
        self, mel_db: np.ndarray, num_iters: int = 60, power: float = 1.0
    ) -> np.ndarray:
        """Converts melspectrogram to waveform using Griffim-Lim"""
        if self.signal_norm:
            mel_db = self.denormalize(mel_db)

        mel_amp = self.db_to_amp(mel_db)
        linear = self.mel_to_linear(mel_amp) ** power

        return self.griffin_lim(linear, num_iters=num_iters)

    def linear_to_mel(self, linear: np.ndarray) -> np.ndarray:
        """Linear spectrogram to mel amp"""
        return np.dot(self._mel_basis, linear)

    def mel_to_linear(self, mel_amp: np.ndarray) -> np.ndarray:
        """Mel amp to linear spectrogram"""
        return np.maximum(1e-10, np.dot(self._inv_mel_basis, mel_amp))

    def amp_to_db(self, mel_amp: np.ndarray) -> np.ndarray:
        return self.spec_gain * np.log10(np.maximum(1e-5, mel_amp))

    def db_to_amp(self, mel_db: np.ndarray) -> np.ndarray:
        return np.power(10.0, mel_db / self.spec_gain)

    # -------------------------------------------------------------------------
    # STFT
    # -------------------------------------------------------------------------
    def stft(self, wav: np.ndarray) -> np.ndarray:
        """Waveform to linear spectrogram"""
        return librosa.stft(
            y=wav,
            n_fft=self.filter_length,
            hop_length=self.hop_length,
            win_length=self.win_length,
            pad_mode="reflect",
        )

    def istft(self, linear: np.ndarray) -> np.ndarray:
        """Linear spectrogram to waveform"""
        return librosa.istft(
            linear, hop_length=self.hop_length, win_length=self.win_length
        )

    def griffin_lim(self, linear: np.ndarray, num_iters: int = 60) -> np.ndarray:
        """Linear spectrogram to waveform using Griffin-Lim"""
        angles = np.exp(2j * np.pi * np.random.rand(*linear.shape))
        linear_complex = np.abs(linear).astype(np.complex)
        audio = self.istft(linear_complex * angles)

        for _ in range(num_iters):
            angles = np.exp(1j * np.angle(self.stft(audio)))
            audio = self.istft(linear_complex * angles)

        return audio

    # -------------------------------------------------------------------------
    # Normalization
    # -------------------------------------------------------------------------

    def normalize(self, mel_db: np.ndarray) -> np.ndarray:
        """Put values in [0, max_norm] or [-max_norm, max_norm]"""
        mel_norm = ((mel_db - self.ref_level_db) - self.min_level_db) / (
            -self.min_level_db
        )
        if self.symmetric_norm:
            # Symmetric norm
            mel_norm = ((2 * self.max_norm) * mel_norm) - self.max_norm
            if self.clip_norm:
                mel_norm = np.clip(mel_norm, -self.max_norm, self.max_norm)
        else:
            # Asymmetric norm
            mel_norm = self.max_norm * mel_norm
            if self.clip_norm:
                mel_norm = np.clip(mel_norm, 0, self.max_norm)

        return mel_norm

    def denormalize(self, mel_db: np.ndarray) -> np.ndarray:
        """Pull values out of [0, max_norm] or [-max_norm, max_norm]"""
        if self.symmetric_norm:
            # Symmetric norm
            if self.clip_norm:
                mel_denorm = np.clip(mel_db, -self.max_norm, self.max_norm)

            mel_denorm = (
                (mel_denorm + self.max_norm) * -self.min_level_db / (2 * self.max_norm)
            ) + self.min_level_db
        else:
            # Asymmetric norm
            if self.clip_norm:
                mel_denorm = np.clip(mel_db, 0, self.max_norm)

            mel_denorm = (
                mel_denorm * -self.min_level_db / self.max_norm
            ) + self.min_level_db

        mel_denorm += self.ref_level_db

        return mel_denorm

    # -------------------------------------------------------------------------
    # Silence Trimming
    # -------------------------------------------------------------------------

    def trim_silence(
        self,
        wav: np.ndarray,
        trim_db: float = 60.0,
        margin_sec: float = 0.01,
        keep_sec: float = 0.1,
    ):
        """
        Trim silent parts with a threshold and margin.
        Keep keep_sec seconds on either side of trimmed audio.
        """
        margin = int(self.sample_rate * margin_sec)
        wav = wav[margin:-margin]
        _, trim_index = librosa.effects.trim(
            wav,
            top_db=trim_db,
            frame_length=self.win_length,
            hop_length=self.hop_length,
        )

        keep_samples = int(self.sample_rate * keep_sec)
        trim_start, trim_end = (
            max(0, trim_index[0] - keep_samples),
            min(len(wav), trim_index[1] + keep_samples),
        )

        return wav[trim_start:trim_end]



"""Utility methods"""


def add_audio_settings(parser):
    """Add audio settings to argparse parser"""
    # STFT settings
    parser.add_argument("--filter-length", type=int, default=1024)
    parser.add_argument("--hop-length", type=int, default=256)
    parser.add_argument("--win-length", type=int, default=1024)
    parser.add_argument("--mel-channels", type=int, default=80)
    parser.add_argument("--sample-rate", type=int, default=22050)
    parser.add_argument("--mel-fmin", type=float, default=0.0)
    parser.add_argument("--mel-fmax", type=float, default=8000.0)
    parser.add_argument("--spec-gain", type=float, default=1.0)

    # Normalization
    parser.add_argument("--ref-level-db", type=float, default=20.0)
    parser.add_argument("--min-level-db", type=float, default=-100.0)
    parser.add_argument("--no-normalize", action="store_true")
    parser.add_argument("--max-norm", type=float, default=4.0)
    parser.add_argument("--asymmetric-norm", action="store_true")
    parser.add_argument("--no-clip-norm", action="store_true")



#!/usr/bin/env python3
"""Command-line interface for wav2mel"""
import argparse
import io
import logging
import os
import sys
from pathlib import Path

import librosa
import numpy as np

from .audio import AudioSettings
from .utils import add_audio_settings

_LOGGER = logging.getLogger("wav2mel")

# -----------------------------------------------------------------------------


def main():
    parser = argparse.ArgumentParser(prog="wav2mel")
    parser.add_argument("wav", nargs="*", help="Path(s) to WAV file(s)")
    parser.add_argument(
        "--batch-dim",
        action="store_true",
        help="Include batch dimension in output arrays",
    )
    parser.add_argument("--output-dir", help="Directory to save numpy file(s)")
    parser.add_argument("--dtype", default="float32", help="numpy data type for mel")

    add_audio_settings(parser)

    # Silence trimming
    parser.add_argument("--trim-silence", action="store_true")
    parser.add_argument("--trim-db", type=float, default=60.0)

    # Miscellaneous
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to the console"
    )
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

        # Disable numba logging
        logging.getLogger("numba").setLevel(logging.WARNING)
    else:
        logging.basicConfig(level=logging.INFO)

    _LOGGER.debug(args)

    if args.output_dir:
        args.output_dir = Path(args.output_dir)
        args.output_dir.mkdir(parents=True, exist_ok=True)
    else:
        args.output_dir = Path.cwd()

    # -------------------------------------------------------------------------

    audio_settings = AudioSettings(
        # STFT
        filter_length=args.filter_length,
        hop_length=args.hop_length,
        win_length=args.win_length,
        mel_channels=args.mel_channels,
        sample_rate=args.sample_rate,
        mel_fmin=args.mel_fmin,
        mel_fmax=args.mel_fmax,
        ref_level_db=args.ref_level_db,
        spec_gain=args.spec_gain,
        #
        # Normalization
        signal_norm=not args.no_normalize,
        min_level_db=args.min_level_db,
        max_norm=args.max_norm,
        clip_norm=not args.no_clip_norm,
        symmetric_norm=not args.asymmetric_norm,
    )

    num_wavs = 0
    if args.wav:
        # Convert to paths
        args.wav = [Path(p) for p in args.wav]

        # Process WAVs
        for wav_path in args.wav:
            try:
                _LOGGER.debug("Processing %s", wav_path)
                wav_array, _ = librosa.load(wav_path, sr=args.sample_rate)
                wav_array = wav_array.astype(args.dtype)

                mel_array = audio_settings.wav2mel(
                    wav_array, trim_silence=args.trim_silence, trim_db=args.trim_db
                )

                if args.batch_dim:
                    mel_array = np.expand_dims(mel_array, 0)

                # Save to numpy file
                mel_path = args.output_dir / ((wav_path.stem) + ".npy")
                np.save(mel_path, mel_array)

                num_wavs += 1
            except Exception:
                _LOGGER.exception(wav_path)
    else:
        # Read from stdin, write to stdout
        if os.isatty(sys.stdin.fileno()):
            print("Reading WAV data from stdin...", file=sys.stderr)

        with io.BytesIO(sys.stdin.buffer.read()) as wav_file:
            wav_array, _ = librosa.load(wav_file, sr=args.sample_rate)

        wav_array = wav_array.astype(args.dtype)

        mel_array = audio_settings.wav2mel(
            wav_array, trim_silence=args.trim_silence, trim_db=args.trim_db
        )

        if args.batch_dim:
            mel_array = np.expand_dims(mel_array, 0)

        # Write to stdout
        with io.BytesIO() as np_file:
            np.save(np_file, mel_array)
            sys.stdout.buffer.write(np_file.getvalue())

        num_wavs += 1

    _LOGGER.info("Done (%s WAV file(s))", num_wavs)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    main()
